/* lmp.h
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef __LMP_H
#define __LMP_H

/* For declaration of asprintf and some others */
#define _GNU_SOURCE

/* load configuration header */
#include <lmp_config.h>
#ifndef DECLSPEC
# ifdef __BEOS__
#  if defined(__GNUC__)
#   define DECLSPEC     __declspec(dllexport)
#  else
#   define DECLSPEC     __declspec(export)
#  endif
# else
# ifdef WIN32
#  define DECLSPEC      __declspec(dllexport)
# else
#  define DECLSPEC
# endif
# endif
#endif

/* c2man cannot cope with 'my' stdlib.h */
#ifndef __C2MAN__
# include <stdlib.h>
# ifdef HAVE_COMPLEX
#  include <complex.h>
# endif
# include <stdlib.h>
#else
#define _Complex
#endif

/* strtold was not prototyped on my linux box which caused an error */
#if defined(__P) && defined(HAVE_STRTOLD)
long double strtold __P ((const char *, char **));
#endif

#define LMP_VERSION          VERSION

#define LMP_ARG_MAX_LENGTH   100
#define LMP_MISSING_ARG      1

#define LMP_NO_ERROR         0
#define LMP_UNKNOWN_OPERATOR 1


/* We will alias the standard value type LMP uses, because
 * not all machines support complex numbers, some even have
 * not implemented long double numbers, so...
 *
 * same with ONE and ZERO
 */

#ifdef HAVE_LONG_DOUBLE
typedef long double lmp_double_t;
#else
typedef double      lmp_double_t;
#endif

#ifdef HAVE_LONG_DOUBLE
# ifdef HAVE_COMPLEX
# define ZERO (lmp_value_type) 0.0l
# define ONE  (lmp_value_type) 1.0l

/* lmp uses as standard type for numbers */
typedef long double __complex__ lmp_value_type;

# else
# define ZERO 0.0l
# define ONE  1.0l

/* lmp uses as standard type for numbers */
typedef long double lmp_value_type;

# endif
#else
# ifdef HAVE_COMPLEX
# define ZERO (lmp_value_type) 0.0
# define ONE  (lmp_value_type) 1.0

/* lmp uses as standard type for numbers */
typedef double _Complex lmp_value_type;

# else
# define ZERO 0.0
# define ONE  1.0

/* lmp uses as standard type for numbers */
typedef double lmp_value_type;

# endif
#endif

typedef lmp_value_type LMP_V;
typedef lmp_double_t   LMP_D;

/* operators lmp knows */
typedef enum lmp_op_enum {
  LMP_OUTER_OR,       // logical or
  LMP_OUTER_AND,      // logical and
  LMP_EQUALS,         // equality
  LMP_UNEQUALS,       // unequality
  LMP_GREATER_EQUAL,  // greater or equal
  LMP_LOWER_EQUAL,    // lower or equal
  LMP_GREATER,        // greater
  LMP_LOWER,          // lower
  LMP_ADDITION,       // +
  LMP_SUBTRACTION,    // -
  LMP_MULTIPLICATION, // *
  LMP_DIVISION,       // /
  LMP_MODULO,         // %, arithmetic remained
  LMP_DIV_NO_REM,     // integer division
  LMP_POWER,          // ^
  LMP_LOG,            // n-base logarithm
  LMP_SINE,           // sine
  LMP_COSINE,         // cosine
  LMP_TANGENT,        // tangent
  LMP_COTANGENT,      // cotangent
  LMP_ARC_SINE,       // arc sine
  LMP_ARC_COSINE,     // arc cosine
  LMP_ARC_TANGENT,    // arc tangent
  LMP_LOGARITHM,      // natural-base logarithm
  LMP_LOG_10,         // 10-base logarithm
  LMP_EXP,            // exponential function
  LMP_SQRT,           // square root
  LMP_UNARY_MINUS     // minus as sign
} lmp_operator;

/* controls are special operations that cannot be simply evaluated. */
typedef enum {
  LMP_DIFF,           // differenziation
  LMP_INT,            // integration
  LMP_SUM,            // summation
  LMP_PROD            // product
} lmp_control;

#define LMP_INT_NO_ROOT (lmp_value_type) -1.0

/* LMP node type enumeration */
typedef enum {
  OP,       // operation like   '+', 'sin', 'log'
  CONTROL,  // control like     'diff', 'sum'
  VAR,      // variable
  VAL       // value
} lmp_node_type;

/* var table entry */
typedef struct {
  char *         var; // Hash identifier
  lmp_value_type val; // Value
} var_table_t;

/* lmp_node_struct alias */
typedef struct lmp_node_struct lmp_node_t;

/* globals */
extern DECLSPEC char **        lmp_args;       // argument vector used by lmp
extern DECLSPEC int            lmp_argnum;     // number of arguments (NEEDS UPDATE)
extern DECLSPEC var_table_t *  lmp_vt;         // var table vector
extern DECLSPEC int            lmp_vt_size;    // var table size
extern DECLSPEC unsigned short lmp_proc_error; // this is used to report errors

/* node that stores operations */
struct lmp_node_operation_struct {
  lmp_operator op;    // operator
  int argc;           // number of arguments
  lmp_node_t ** argv; // argument vector
};

/* node that stores controls */
struct lmp_node_control_struct {
  lmp_control op;     // control operator
  int argc;           // number of arguments
  lmp_node_t ** argv; // argument vector
};

/* node to store trees */
struct lmp_node_struct {
  lmp_node_type what;                            // what a node is this one
  union {                                        // could be:
    struct lmp_node_operation_struct operation;  //   operation
    struct lmp_node_control_struct control;      //   control
    lmp_value_type value;                        //   value
    char * variable;                             //   variable
  } data;
};

typedef struct lmp_stack_struct lmp_stack_t;

struct lmp_stack_struct {
  lmp_node_t ** sp;
  int si;
  int bs;
};

typedef struct lmp_simpl_struct lmp_simpl_t;

struct lmp_simpl_struct {
  unsigned int do_balance : 1;
  unsigned int do_level1  : 1;
  unsigned int do_level2  : 1;
  unsigned int do_unary_minus : 1;
};

#ifdef __cplusplus
extern "C" {
#endif

/* LMP parse functions */
extern DECLSPEC lmp_node_t * lmp_parse_braces ();       // Parse entities and subexpressions (in braces)
extern DECLSPEC lmp_node_t * lmp_parse_power ();        // Parse power-level operations
extern DECLSPEC lmp_node_t * lmp_parse_diff ();         // Parse differenzation-level operations
extern DECLSPEC lmp_node_t * lmp_parse_trigonometry (); // Parse trigonometry operations
extern DECLSPEC lmp_node_t * lmp_parse_logarithm ();    // Parse (n-base) logarithm-level operations
extern DECLSPEC lmp_node_t * lmp_parse_modulo ();       // Parse modulo-level operations
extern DECLSPEC lmp_node_t * lmp_parse_mul ();          // Parse *-level operations
extern DECLSPEC lmp_node_t * lmp_parse_sum ();          // Parse sum and prod controls
extern DECLSPEC lmp_node_t * lmp_parse_addition ();     // Parse addition-level operations
extern DECLSPEC lmp_node_t * lmp_parse_compare ();      // Parse comparatitive operations
extern DECLSPEC lmp_node_t * lmp_parse_and ();          // Parse and-level operations
extern DECLSPEC lmp_node_t * lmp_parse ();              // Parse expressions

/* Return the version of the LMP library */
extern DECLSPEC char * lmp_ver (); // Return LMP version string

/* LMP message functions (NEED UPDATE) */
extern DECLSPEC void lmp_stop (const char * msg);
extern DECLSPEC void lmp_error (const char * msg, ...);
extern DECLSPEC void lmp_warning (const char * msg, ...);

/* LMP scanner functions */
/* a) LMP input string scanning functions */
extern DECLSPEC char ** lmp_iss_by_space (const char * string, int * argc); // scan string by space characters
extern DECLSPEC char ** lmp_iss_by_elements (char * string, int * argc);    // scan string by known elements

/* b) LMP token functions */
extern DECLSPEC char * lmp_token_next ();           // returns next token
extern DECLSPEC LMP_V  lmp_token_next_number ();    // returns next number
extern DECLSPEC LMP_V  lmp_token_next_complex ();   // returns next complex value
extern DECLSPEC int    lmp_token_next_is_number (); // returns TRUE if the next token is a number
extern DECLSPEC char * lmp_token_prev ();           // returns previous token

extern DECLSPEC char * lmp_token_following ();      // return next token, but does not increment index
extern DECLSPEC int    lmp_token_following_equals (char *s); // compares next token with s

extern DECLSPEC int    lmp_token_inc ();            // increment token index
extern DECLSPEC int    lmp_token_args ();           // returns number of tokens
extern DECLSPEC int    lmp_token_is_last ();        // Any tokens waiting?

/* LMP var table functions */
extern DECLSPEC int lmp_vt_lookup (const char * s, lmp_value_type * val); // look up an identifier in the var table
extern DECLSPEC int lmp_vt_init   (const char * vs);                      // init var table with vs
extern DECLSPEC int lmp_vt_add    (const char * var, lmp_value_type val); // add entry var=val to var table

/* Evaluate a tree */
extern DECLSPEC lmp_value_type lmp_eval (lmp_node_t * tree);

/* Compute a control */
extern DECLSPEC lmp_node_t *   lmp_proc_control (lmp_node_t * control);

/* Compute a binary operation */
extern DECLSPEC lmp_value_type lmp_proc_binary (lmp_operator op, lmp_value_type left, lmp_value_type right);

/* Compute a unary operation */
extern DECLSPEC lmp_value_type lmp_proc_unary (lmp_operator op, lmp_value_type operand);

/* LMP conversion functions */
/* convert base to another */
extern DECLSPEC char * lmp_conv_base (unsigned base, long value, int precision);
extern DECLSPEC char * lmp_conv_base_unsigned (unsigned base, unsigned long value, int precision);

/* LMP node functions */
/* Copy a node */
extern DECLSPEC lmp_node_t * lmp_node_copy (lmp_node_t * n /* input node */);
/* Create a new node */
extern DECLSPEC lmp_node_t * lmp_node_new (lmp_node_type nt /* what type of a node */, ...);
/* Create a new node */
extern DECLSPEC lmp_node_t * lmp_node_new_fp_safe (lmp_node_type nt, lmp_value_type value);
/* Free a node */
extern DECLSPEC int          lmp_node_free (lmp_node_t * root);

/* a node is a factored variable */
extern DECLSPEC int          lmp_node_factored (lmp_node_t *node, lmp_node_t *var, lmp_node_t **fact);
/* a node is a polynom fragment: a*x^b */
extern DECLSPEC int          lmp_node_polynom_fragment (lmp_node_t *node, lmp_node_t *var, lmp_node_t **fact, lmp_node_t **exp);
/* a node is a polynom */
extern DECLSPEC int          lmp_node_polynom (lmp_node_t *node, lmp_node_t *var, lmp_stack_t *fact, lmp_stack_t *exp);
/* creates a polynom fragment */
extern DECLSPEC lmp_node_t * lmp_node_create_polynom_fragment (lmp_node_t *var, lmp_node_t *fact, lmp_node_t *exp);

/* Simplify a tree */

extern DECLSPEC lmp_node_t * lmp_tree_simplify_full (lmp_node_t * root /* input node */);

extern DECLSPEC lmp_node_t * lmp_tree_simplify (lmp_node_t * root /* input node */, lmp_simpl_t simpl /* Simplification level */);

extern DECLSPEC lmp_node_t * lmp_tree_simplify_level1 (lmp_node_t * root /* input node */);
extern DECLSPEC lmp_node_t * lmp_tree_simplify_level2 (lmp_node_t * root /* input node */);
extern DECLSPEC lmp_node_t * lmp_tree_simplify_unary_minus (lmp_node_t * root /* input node */);
/* create a more balanced tree */
extern DECLSPEC lmp_node_t * lmp_tree_balance  (lmp_node_t * root);
/* check if two trees are equal */
extern DECLSPEC int          lmp_tree_equals   (lmp_node_t * a, lmp_node_t * b);
/* replace a var by its value */
extern DECLSPEC lmp_node_t * lmp_tree_var_eval (lmp_node_t * var, lmp_node_t * root);
/* returns TRUE if tree contains var */
extern DECLSPEC int          lmp_tree_contains (lmp_node_t * tree, lmp_node_t * var); 
/* returns TRUE if the tree can be evaluated (no free variables) */
extern DECLSPEC int          lmp_tree_evaluable (lmp_node_t * tree);

/* LMP controls */
/* Control function for differenziation */
extern DECLSPEC lmp_node_t * lmp_ctrl_diff (lmp_node_t * var, lmp_node_t * root);
/* Control function for integration */
extern DECLSPEC lmp_node_t * lmp_ctrl_int (lmp_node_t * var, lmp_node_t * root);
/* Control function for sums */
extern DECLSPEC lmp_node_t * lmp_ctrl_sum (lmp_node_t * var, lmp_node_t * start, lmp_node_t * end, lmp_node_t * root);
/* Control function for prods */
extern DECLSPEC lmp_node_t * lmp_ctrl_prod (lmp_node_t * var, lmp_node_t * start, lmp_node_t * end, lmp_node_t * root);

/* LMP infix functions */
extern DECLSPEC char * lmp_infix_from_tree (lmp_node_t * root, int expand); // Convert a tree into infix
extern DECLSPEC char * lmp_infix_op2str    (lmp_operator op);   // Look up the name of an operator
extern DECLSPEC char * lmp_infix_ctrl2str  (lmp_control  ctrl); // Look up the name of a control

/* LMP output functions */
extern DECLSPEC char * lmp_output_latex (lmp_node_t * root, int expand); // Convert a tree into infix
extern DECLSPEC char * lmp_output_op2latex (lmp_operator op);   // Look up the name of an operator
extern DECLSPEC char * lmp_output_ctrl2latex (lmp_control ctrl); // Look up the name of a control

#ifndef HAVE_ASPRINTF
extern void          asprintf (char ** in, const char * format, ...);
#endif
/* print a tree */
extern DECLSPEC void lmp_print_tree    (const char * lead, const char * add, lmp_node_t * root);
extern DECLSPEC void lmp_print_tree_pc (const char * lead, const char * add, lmp_node_t * root);
extern DECLSPEC void lmp_print_number  (lmp_double_t val);
extern DECLSPEC char * lmp_print_number_to_string  (lmp_double_t val);

extern DECLSPEC lmp_node_t *  lmp_stack_get   (lmp_stack_t * stack, int n);
extern DECLSPEC int           lmp_stack_push  (lmp_stack_t * stack, lmp_node_t * node);
extern DECLSPEC lmp_stack_t * lmp_stack_new   ();
extern DECLSPEC lmp_stack_t * lmp_stack_copy  (lmp_stack_t * stack);
extern DECLSPEC int           lmp_stack_free  (lmp_stack_t * stack);
extern DECLSPEC int           lmp_stack_split (lmp_stack_t * stack, lmp_stack_t ** l, lmp_stack_t ** r);


extern DECLSPEC void          lmp_stack_dump  (lmp_stack_t * stack);

#ifdef __cplusplus
}
#endif
#endif
