/* lmpmain.c
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include "lmp.h"
#include "lmpinput.h"
#include "lmpmath.h"
#include "lmphelp.h"

#define MAX_FORMAT_BUF  255

#define CONVB(s,b,v,p) s ? lmp_conv_base (b,v,p) : lmp_conv_base_unsigned (b,v,p)

/* Format a printf-formatter string */
char * lmp_conv_dyn_format (start, value1, value2, end)
const char * start; /* pre string*/
int value1;         /* precision */
int value2;         /* significance */
const char * end;   /* post string */
{
  static char result [MAX_FORMAT_BUF + 1];

  if (value1 != -1 && value2 == -1)
    sprintf (result, "%s%%.%dLf%s", start, value1, end);
  else
    sprintf (result, "%s%%.%dLg%s", start, value2, end);


  return result;
}

enum {
  STANDARD,
  LATEX
};

/* LMP main function
 *
 * It parses command line arguments and calls the L math
 * processor.
 * returns 0 if successfull.
 */
int main (argc, argv)
int argc;     /* Number of arguments */
char ** argv; /* Vector of arguments */
{
  lmp_value_type result;
  lmp_node_t * root;

  lmp_simpl_t simplify = { 1, 1, 1, 1 };
  int i, lmp_argc,
    base = -1, 
    lmp_offset = 0, 
    sig_prec=-1, 
    prec=-1, 
    parse_args = 1,
    dump_init = 1,
    computing = 1,
    print_tree = 0,
    signed_base_conv = 1,
    output_style = STANDARD;

  char * endptr[1];
  char ** env_argv;
  int     env_argc;

  if (argc == 1)
    lmp_print_usage ();

  if (!strcmp (argv[1], "--help"))
    lmp_print_help ();

  if (!strcmp (argv[1], "--version"))
    lmp_print_version ();

  /* Initialize from environment variable LMP_OPTIONS
   * You may export this variable on the shell to set
   * default behavior of LMP, e.g:
   *
   * LMP_OPTIONS='-pa -qi -p 10'
   *
   * - parse the arguments
   * - silent initialization
   * - 10 decimals after the point
   *
   * You may override this when calling lmp.
   */
  env_argv = lmp_iss_by_space (getenv ("LMP_OPTIONS"), &env_argc);

  for (i = 0; i < env_argc; i++) {
    if (!strcmp (env_argv[i], "-pa") || !strcmp (env_argv[i], "--parse-args")) {
      parse_args=1;
    }
    else if (!strcmp (env_argv[i], "-np") || !strcmp (env_argv[i], "--no-parse-args")) {
      parse_args=1;
    }
    else if (!strcmp (env_argv[i], "-qi") || !strcmp (env_argv[i], "--quiet-init")) {
      dump_init = 0;
    }
    else if (!strcmp (env_argv[i], "-vi") || !strcmp (env_argv[i], "--verbose-init")) {
      dump_init = 1;
    }
    else if (!strcmp (env_argv [i], "-ns") || !strcmp (env_argv [i], "--no-simplify")) {
      simplify.do_balance = 0;
      simplify.do_level1  = 0;
      simplify.do_level2  = 0;
      simplify.do_unary_minus  = 0;
    }
    else if (!strcmp (env_argv [i], "-si") || !strcmp (env_argv [i], "--simplify")) {
      simplify.do_balance = 1;
      simplify.do_level1  = 1;
      simplify.do_level2  = 1;
      simplify.do_unary_minus  = 1;
    }
    else if (!strcmp (env_argv [i], "-ps") || !strcmp (env_argv [i], "--post-simplify")) {
      simplify.do_level2  = 1;
    }
    else if (!strcmp (env_argv [i], "-nps") || !strcmp (env_argv [i], "--no-post-simplify")) {
      simplify.do_level2  = 0;
    }
    else if (!strcmp (env_argv [i], "-ums") || !strcmp (env_argv [i], "--unary-minus-simplify")) {
      simplify.do_unary_minus = 1;
    }
    else if (!strcmp (env_argv [i], "-nums") || !strcmp (env_argv [i], "--no-unary-minus-simplify")) {
      simplify.do_unary_minus = 0;
    }
    else if (!strcmp (env_argv [i], "-bal") || !strcmp (env_argv [i], "--no-balancing")) {
      simplify.do_balance = 0;
    }
    else if (!strcmp (env_argv [i], "-nb") || !strcmp (env_argv [i], "--balancing")) {
      simplify.do_balance = 1;
    }
    else if (!strcmp (env_argv [i], "-jb") || !strcmp (env_argv [i], "--just-balancing")) {
      simplify.do_balance = 1;
      simplify.do_level1  = 0;
      simplify.do_level2  = 0;
      simplify.do_unary_minus  = 0;
    }
    else if (!strcmp (env_argv [i], "-ub") || !strcmp (env_argv [i], "--unsigned-base-output")) {
      signed_base_conv = 0;
    }
    else if (!strcmp (env_argv [i], "-sb") || !strcmp (env_argv [i], "--signed-base-output")) {
      signed_base_conv = 1;
    }
    else if (!strcmp (env_argv[i], "-b")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      base = STRTOVAL (env_argv[++i], endptr, 10);
      if (**endptr != '\0' || *(env_argv [i]) == '\0') {
	fprintf (stderr, "Base argument must be a number, treating as 10.\n");
	base = 10;
      }
    }
    else if (!strncmp (env_argv[i], "--base=", 7)) {
      base = STRTOVAL (env_argv[i]+7, endptr, 10);
      if (**endptr != '\0' || *(env_argv [i]+7) == '\0') {
	fprintf  (stderr, "Base argument must be a number, treating as 10.\n");
	base = 10;
      }
    }
    else if (!strcmp (env_argv[i], "-sp")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      sig_prec = STRTOVAL (env_argv[++i], endptr, 10);
      if (**endptr != '\0' || *(env_argv [i]) == '\0') {
	fprintf (stderr, "Significant precision argument must be a number, treating as 6.\n");
	sig_prec = 6;
      }
    }
    else if (!strncmp (env_argv[i], "--significance=", 15)) {
      sig_prec = STRTOVAL (env_argv[i]+15, endptr, 10);
      if (**endptr != '\0' || *(env_argv [i]+15) == '\0') {
	fprintf  (stderr, "Significant precision argument must be a number, treating as 6.\n");
	sig_prec = 6;
      }
    }
    else if (!strcmp (env_argv[i], "-p")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      prec = STRTOVAL (env_argv[++i], endptr, 10);
      if (**endptr != '\0' || *(argv [i]) == '\0') {
	fprintf (stderr, "Precision argument must be a number, treating as 6.\n");
	prec = 6;
      }
    }
    else if (!strncmp (env_argv[i], "--precision=", 12)) {
      prec = STRTOVAL (env_argv[i]+12, endptr, 10);
      if (**endptr != '\0' || *(argv [i]+12) == '\0') {
	fprintf  (stderr, "Precision argument must be a number, treating as 6.\n");
	prec = 6;
      }
    }
    else if (!strcmp (env_argv[i], "-vt")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      lmp_vt_init (env_argv[++i]);
    }
    else if (!strncmp (env_argv[i], "--vartable=", 11)) {
      lmp_vt_init (&env_argv[i][11]);
    }
    else if (!strcmp (env_argv [i], "--latex") || !strcmp (env_argv [i], "-l")) {
      output_style = LATEX;
    }
  }

  for (i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-pa") || !strcmp (argv[i], "--parse-args")) {
      parse_args=1;
      lmp_offset++;
    }
    else if (!strcmp (argv[i], "-np") || !strcmp (argv[i], "--no-parse-args")) {
      parse_args=0;
      lmp_offset++;
    }
    else if (!strcmp (argv[i], "-ub") || !strcmp (argv[i], "--unsigned-base-output")) {
      signed_base_conv = 0;
      lmp_offset++;
    }
    else if (!strcmp (argv[i], "-sb") || !strcmp (argv[i], "--signed-base-output")) {
      signed_base_conv = 1;
      lmp_offset++;
    }
    else if (!strcmp (argv[i], "-qi") || !strcmp (argv[i], "--quiet-init")) {
      lmp_offset++;
      dump_init = 0;
    }
    else if (!strcmp (argv[i], "-vi") || !strcmp (argv[i], "--verbose-init")) {
      lmp_offset++;
      dump_init = 1;
    }
    else if (!strcmp (argv[i], "-b")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      lmp_offset += 2;
      base = STRTOVAL (argv[++i], endptr, 10);
      if (**endptr != '\0' || *(argv[i]) == '\0') {
	fprintf (stderr, "Base argument must be a number, treating as 10.\n");
	base = 10;
      }
    }
    else if (!strncmp (argv[i], "--base=", 7)) {
      lmp_offset++;
      base = STRTOVAL (argv[i]+7, endptr, 10);
      if (**endptr != '\0' || *(argv [i]+7) == '\0') {
	fprintf  (stderr, "Base argument must be a number, treating as 10.\n");
	base = 10;
      }
    }
    else if (!strcmp (argv[i], "-sp")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      lmp_offset += 2;
      sig_prec = STRTOVAL (argv[++i], endptr, 10);
      if (**endptr != '\0' || *(argv [i]) == '\0') {
	fprintf (stderr, "Significant precision argument must be a number, treating as 6.\n");
	sig_prec = 6;
      }
    }
    else if (!strncmp (argv[i], "--significance=", 15)) {
      lmp_offset++;
      sig_prec = STRTOVAL (argv[i]+15, endptr, 10);
      if (**endptr != '\0' || *(argv [i]+15) == '\0') {
	fprintf  (stderr, "Significant precision argument must be a number, treating as 6.\n");
	sig_prec = 6;
      }
    }
    else if (!strcmp (argv[i], "-p")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      lmp_offset += 2;
      prec = STRTOVAL (argv[++i], endptr, 10);
      if (**endptr != '\0' || *(argv [i]) == '\0') {
	fprintf (stderr, "Precision argument must be a number, treating as 6.\n");
	prec = 6;
      }
    }
    else if (!strncmp (argv[i], "--precision=", 12)) {
      lmp_offset++;
      prec = STRTOVAL (argv[i]+12, endptr, 10);
      if (**endptr != '\0' || *(argv [i]+12) == '\0') {
	fprintf  (stderr, "Precision argument must be a number, treating as 6.\n");
	prec = 6;
      }
    }
    else if (!strcmp (argv[i], "-vt")) {
      if (i == argc-1) {
	fprintf (stderr, "No idea what you want, exiting.\n");
	exit (0);
      }
      lmp_offset += 2;
      lmp_vt_init (argv[++i]);
    }
    else if (!strncmp (argv[i], "--vartable=", 11)) {
      lmp_offset++;
      lmp_vt_init (&argv[i][11]);
    }
    else if (!strcmp (argv[i], "-qi") || !strcmp (argv[i], "--quiet-init")) {
      lmp_offset++;
      dump_init = 0;
    }
    else if (!strcmp (argv [i], "-nc") || !strcmp (argv [i], "--no-computing")) {
      lmp_offset++;
      computing = 0;
    }
    else if (!strcmp (argv [i], "-ns") || !strcmp (argv [i], "--no-simplify")) {
      lmp_offset++;
      simplify.do_balance = 0;
      simplify.do_level1  = 0;
      simplify.do_level2  = 0;
      simplify.do_unary_minus = 0;
    }
    else if (!strcmp (argv [i], "-si") || !strcmp (argv [i], "--simplify")) {
      lmp_offset++;
      simplify.do_balance = 1;
      simplify.do_level1  = 1;
      simplify.do_level2  = 1;
      simplify.do_unary_minus = 1;
    }
    else if (!strcmp (argv [i], "-ps") || !strcmp (argv [i], "--post-simplify")) {
      lmp_offset++;
      simplify.do_level2  = 1;
    }
    else if (!strcmp (argv [i], "-ums") || !strcmp (argv [i], "--unary-minus-simplify")) {
      lmp_offset++;
      simplify.do_unary_minus = 1;
    }
    else if (!strcmp (argv [i], "-bal") || !strcmp (argv [i], "--balancing")) {
      lmp_offset++;
      simplify.do_balance  = 1;
    }
    else if (!strcmp (argv [i], "-jb") || !strcmp (argv [i], "--just-balancing")) {
      lmp_offset++;
      simplify.do_balance = 1;
      simplify.do_level1  = 0;
      simplify.do_level2  = 0;
      simplify.do_unary_minus = 0;
    }
    else if (!strcmp (argv [i], "-nb") || !strcmp (argv [i], "--no-balancing")) {
      lmp_offset++;
      simplify.do_balance = 0;
    }
    else if (!strcmp (argv [i], "-nps") || !strcmp (argv [i], "--no-post-simplify")) {
      lmp_offset++;
      simplify.do_level2 = 0;
    }
    else if (!strcmp (argv [i], "-nums") || !strcmp (argv [i], "--no-unary-minus-simplify")) {
      lmp_offset++;
      simplify.do_unary_minus = 0;
    }
    else if (!strcmp (argv [i], "-pt") || !strcmp (argv [i], "--print-tree")) {
      lmp_offset++;
      print_tree = 1;
    }
    else if (!strcmp (argv [i], "--latex") || !strcmp (argv [i], "-l")) {
      lmp_offset++;
      output_style = LATEX;
    }
  }

  /* Two methods are available:
   * - Parse each argument
   * - Take arguments as a parsed string
   */
  if (parse_args) { // Parse each argument, new method
    char ** args_bak;
    for (i = lmp_offset + 1; i < argc; i++) {
      int init; // Mark init mode

      args_bak = lmp_args = lmp_iss_by_elements (argv[i], &lmp_argc);

      init = 0;
      if (lmp_argc > 1) {
	if (!strcmp (lmp_args[1], "=")) {
	  lmp_args = &lmp_args [2];
	  root = lmp_parse ();
	  lmp_vt_add (args_bak [0], (result = lmp_eval (root)));
	  init = 1;
	}
      }

      if (!init || dump_init) { // Init quiet-init mode when won't show results
	if (!init) {
	  lmp_node_t * p;

	  lmp_argnum = 1;
	  root = lmp_parse ();
	  p = root;
	  root = lmp_tree_simplify (root, simplify);
	  lmp_node_free (p);
	}
	if (print_tree)
	  lmp_print_tree_pc ("", "  ", root);
	if (computing) {
	  if (!init) {
	    result = lmp_eval (root);
	  }
	  lmp_node_free (root);
	  if (!lmp_token_is_last ())
	    lmp_error ("Nothing expected at argument %d: %s\n", lmp_token_args (), lmp_token_next ());
	  if (IS_REAL (result)) {
	    if (base == -1 && sig_prec == -1 && prec == -1) {
	      lmp_print_number (CREAL (result));
	      fputc ('\n', stdout);
	    }
	    else if (base >= 2 && base <= 36)
	      printf ("%s\n", CONVB (signed_base_conv, base, result, prec));
	    else if (sig_prec == -1 && prec == -1)
	      fprintf (stderr, "Illegal base specification: %d\n", base);
	    else
	      printf (lmp_conv_dyn_format ("", prec, sig_prec, "\n"), result);
	  }
	  else { // if complex
	    if (base == -1 && sig_prec == -1 && prec == -1) {
	      lmp_print_number (CREAL (result));
	      printf (" + ");
	      lmp_print_number (CIMAG (result));
	      printf ("i\n");
	    }
	    else if (base >= 2 && base <= 36) {
	      printf ("%s + ", CONVB (signed_base_conv, base, CREAL (result), prec));
	      printf ("%si\n", CONVB (signed_base_conv, base, CIMAG (result), prec));
	    }
	    else if (sig_prec == -1 && prec == -1)
	      fprintf (stderr, "Illegal base specification: %d\n", base);
	    else {
	      printf (lmp_conv_dyn_format ("", prec, sig_prec, " + "), CREAL (result));
	      printf (lmp_conv_dyn_format ("", prec, sig_prec, "i\n"), CIMAG (result));
	    }
	  }
	}
	else {
	  if (output_style == STANDARD)
	    printf ("%s\n", lmp_infix_from_tree (root, 1));
	  else
	    printf ("%s = %s\n", lmp_output_latex (root, 0), lmp_output_latex (root, 1));

	  free (root);
	}
      }
    }
  }
  else { // Arguments are parsed already
    lmp_node_t * p;

    lmp_args = &argv [lmp_offset+1];
    p = root = lmp_parse ();
    root = lmp_tree_simplify (root, simplify);
    lmp_node_free (p);

    if (print_tree)
      lmp_print_tree_pc ("", "  ", root);

    if (computing) {
      result = lmp_eval (root);
      free (root);
      if (lmp_token_args () != argc - lmp_offset)
	lmp_error ("Nothing expected at argument %d\n", lmp_offset);
      if (IS_REAL (result)) {
	if (base == -1 && sig_prec == -1 && prec == -1) {
	  lmp_print_number (CREAL (result));
	  fputc ('\n', stdout);
	}
	else if (base >= 2 && base <= 36) {
	  printf ("%s\n", CONVB (signed_base_conv, base, CREAL (result), prec));
	}
	else if (sig_prec == -1 && prec == -1)
	  fprintf (stderr, "Illegal base specification: %d\n", base);
	else {
	  printf (lmp_conv_dyn_format ("", prec, sig_prec, "\n"), CREAL (result));
	}
      }
      else { // if complex
	if (base == -1 && sig_prec == -1 && prec == -1) {
	  lmp_print_number (CREAL (result));
	  printf (" + ");
	  lmp_print_number (CIMAG (result));
	  printf ("i\n");
	}
	else if (base >= 2 && base <= 36) {
	  printf ("%s + ", CONVB (signed_base_conv, base, CREAL (result), prec));
	  printf ("%si\n", CONVB (signed_base_conv, base, CIMAG (result), prec));
	}
	else if (sig_prec == -1 && prec == -1)
	  fprintf (stderr, "Illegal base specification: %d\n", base);
	else {
	  printf (lmp_conv_dyn_format ("", prec, sig_prec, " + "), CREAL (result));
	  printf (lmp_conv_dyn_format ("", prec, sig_prec, "i\n"), CIMAG (result));
	}
      }
    }
    else {
      if (output_style == STANDARD)
	printf ("%s\n", lmp_infix_from_tree (root, 1));
      else
	printf ("%s = %s\n", lmp_output_latex (root, 0), lmp_output_latex (root, 1));
    }
  }

  return 0;
}
