/* lmpproc.c
 *
 * Copyright (C) 1999-2000 Bernhard Pietsch <bjtp@gmx.net>
 *
 * This file is part of LMP library.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include "lmp.h"
#include "lmpmath.h"

lmp_node_t * lmp_proc_control (lmp_node_t * control) {
  lmp_node_t * res;

  res = NULL;
  switch (control->data.control.op) {
  case LMP_DIFF:
    res = lmp_ctrl_diff (control->data.control.argv [0], control->data.control.argv [1]);
    break;
  case LMP_INT:
    res = lmp_ctrl_int (control->data.control.argv [0], control->data.control.argv [1]);
    break;
  case LMP_SUM:
    res = lmp_ctrl_sum (control->data.control.argv [0],
			control->data.control.argv [1],
			control->data.control.argv [2],
			control->data.control.argv [3]);
    break;
  case LMP_PROD:
    res = lmp_ctrl_prod (control->data.control.argv [0],
			 control->data.control.argv [1],
			 control->data.control.argv [2],
			 control->data.control.argv [3]);
    break;
  }

  return res;
}

/* Binary function processor
 *
 * This will compute all math functions with two operands.
 * returns the result of the operation
 */
lmp_value_type lmp_proc_binary (op, left, right)
lmp_operator op;      /* operation */
lmp_value_type left;  /* left operand */
lmp_value_type right; /* right operand */
{
  switch (op) {
  case LMP_OUTER_OR:
    return left || right;

  case LMP_OUTER_AND:
    return left && right;

  case LMP_EQUALS:
    return left == right;

  case LMP_UNEQUALS:
    return left != right;

  case LMP_GREATER_EQUAL:
    return CREAL (left) >= CREAL (right);

  case LMP_LOWER_EQUAL:
    return CREAL (left) <= CREAL (right);

  case LMP_GREATER:
    return CREAL (left) > CREAL (right);

  case LMP_LOWER:
    return CREAL (left) < CREAL (right);

  case LMP_ADDITION:
    return left + right;

  case LMP_SUBTRACTION:
    return left - right;

  case LMP_MULTIPLICATION:
    return left * right;

  case LMP_DIVISION:
    return left / right;

  case LMP_MODULO:
    return MOD (left, right);

  case LMP_DIV_NO_REM:
    return (long long) left / (long long) right;

  case LMP_POWER:
#ifdef HAVE_COMPLEX
    if (!IS_REAL (left) || !IS_REAL (right) || ((CREAL (left) < 0) && (!IS_INT (CREAL (right)))))
      return CPOW (left, right);
    else
#endif
      return POW (left, right);

  case LMP_LOG:
#ifdef HAVE_COMPLEX
    if (IS_REAL (left) && IS_REAL (right))
      return LOGN (left) / LOGN (right);
    else
      return CLOGN (left) / CLOGN (right);
#else
    return LOGN (left) / LOGN (right);
#endif
    

  default:
    lmp_proc_error = LMP_UNKNOWN_OPERATOR;
  }

  return 0;
}

/* Unary function processor
 *
 * This will compute all math functions with one operand.
 * returns the result of the operation
 */
lmp_value_type lmp_proc_unary (op, operand)
     lmp_operator op;         /* operation */
     lmp_value_type operand;  /* operand */
{
#ifdef HAVE_COMPLEX
  if (IS_REAL (operand)) {
#endif
    switch (op) {
    case LMP_SINE:
      return SIN (operand);

    case LMP_COSINE:
      return COS (operand);

    case LMP_TANGENT:
      return TAN (operand);

    case LMP_COTANGENT:
      return 1.0l / TAN (operand);

    case LMP_ARC_SINE:
      return ARCSIN (operand);

    case LMP_ARC_COSINE:
      return ARCCOS (operand);

    case LMP_ARC_TANGENT:
      return ARCTAN (operand);

    case LMP_LOGARITHM:
      return LOGN (operand);

    case LMP_LOG_10:
      return LOG10 (operand);

    case LMP_EXP:
      return EXP (operand);

    case LMP_SQRT:
#ifdef HAVE_COMPLEX
      if (CREAL (operand) >= 0)
	return SQRT (operand);
      else
	return CSQRT (operand);
#else
      return SQRT (operand);
#endif

    case LMP_UNARY_MINUS:
      return - operand;

    default:
      lmp_proc_error = LMP_UNKNOWN_OPERATOR;
    }
#ifdef HAVE_COMPLEX
  }
  else {
    switch (op) {
    case LMP_SINE:
      return CSIN (operand);

    case LMP_COSINE:
      return CCOS (operand);

    case LMP_TANGENT:
      return CTAN (operand);

    case LMP_COTANGENT:
      return 1.0l / CTAN (operand);

    case LMP_ARC_SINE:
      return CARCSIN (operand);

    case LMP_ARC_COSINE:
      return CARCCOS (operand);

    case LMP_ARC_TANGENT:
      return CARCTAN (operand);

    case LMP_LOGARITHM:
      return CLOGN (operand);

    case LMP_LOG_10:
      return CLOG10 (operand);

    case LMP_EXP:
      return CEXP (operand);

    case LMP_SQRT:
      return CSQRT (operand);

    case LMP_UNARY_MINUS:
      return - operand;

    default:
      lmp_proc_error = LMP_UNKNOWN_OPERATOR;
    }    
  }
#endif
  return 0;
}
